//
// GameX -- Missile-Defender Demo
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#define GAMEX_MAIN
#pragma comment (lib, "gamex.lib")
#include "gamex.hpp"
										//*** Global Variables & Defines
#define NUM_CITIES		7				// Number of Cities we'll have
#define MAX_MISSILES  512				// Maximum missiles we'll allow

#define	GROUND_FRACTION	0.1f // make 10% of the screen the ground
#define	SKY_FRACTION	(1.0f-GROUND_FRACTION) // and the rest the sky

float gravity = 0.03f; // arbitrary gravitational acceleration constant
float fadingIn = 1.0f; // 1.0 for blacked-out screen, 0.0 for normal
bool paused = false; // true if game is paused, false otherwise
int screen_width, screen_height, ground_pos; // pixel dimension values
ColorX skyColor, groundColor, turretColor; // gradient colors
ImageX city_img, crosshair_img, explode_img[3]; // various Images

struct City {
	int x, y; // x and y position
	int radius; // 1.0 = normal size
	float alive; // 1.0 = alive, < 1.0 = dying, -0.25 = dead
	float charged; // 1.0 = charged, < 1.0 = recharging
	int turret_x, turret_y; // turret extent and direction
	ColorX color; // (solid) shading color of this city (has no actual gameplay effect)
	int explode_img_id; // 0 means color is red-ish, 1 for green-ish, 2 for blue-ish
};

City city [NUM_CITIES]; // array of all the ground cities we'll have

struct Missile {
	float x, y; // x and y position
	float dx, dy; // x and y velocities
	int origin; // -1 for enemy, other # for a city of origin
	float exploding; // 0.0 for normal missile, > 0.0 for exploding missile, value indicates current blast radius
	float angle; // amgle at which this missile's explosion will be drawn
	float max_explode_radius; // radius that this missile's explosion will expand to
	int img_id; // 0 means color is red-ish, 1 for green-ish, 2 for blue-ish
};

Missile missile [MAX_MISSILES]; // array of all the missiles we can possibly have at once
int num_missiles = 0; // number of missiles actually being used at the moment


// Custom function declarations:

// Functions to be called from GameInit():
void InitGameX (void), LoadImages (void), PrepareColors (void), InitCities (void);

// Functions to be called from GameRun():
void PositionMouse (void),PositionTurrets (void), UpdateCities (void), LaunchMissiles (void);
void UpdateMissiles (void), UpdateFade (void), UpdatePause (void);

// Functions to be called from GameDraw():
void DrawCrosshair (void), DrawGround (void), DrawCities (bool live);
void DrawMissiles (void), DrawExplosions (void), DrawFade (void);

// Other helper functions:
int FindNearestCity (int x, bool must_be_charged);


// Required GameInit function:
void GameInit (void)
{
	InitGameX(); // Initialize GameX itself (inside this function)
	LoadImages(); // Load the images we want to use
	PrepareColors(); // Set up a few colors that would be annoying to constantly re-create later
	InitCities(); // Initialize the cities (position, color, size, health)
}

// Required GameRun function:
void GameRun (void)
{
	UpdatePause(); // pause/unpause if P was pressed
	UpdateFade(); // slowly unfade game if it's just starting up
	if(paused) return; // if paused, don't update rest of game state
	PositionMouse(); // keyboard-move the mouse and limit it to firing range
	PositionTurrets(); // point the city turrets at the mouse
	UpdateCities(); // recharge weapons and update status of cities
	LaunchMissiles(); // launch missiles if input requests it, and launch enemy missiles at set interval
	UpdateMissiles(); // move missiles and explosions and do collision detection
}

// Required GameDraw function:
void GameDraw (void)
{
	GameX.FillScreen(skyColor); // draw the sky background
	DrawMissiles(); // draw all non-exploding missiles
	DrawCities(false); // draw dying cities BEFORE explosions so explosions draw on top of them
	DrawExplosions(); // draw all exploding missiles
	DrawCities(true); // draw active cities AFTER explosions so explosions draw behind them
	DrawGround(); // draw the ground at the bottom of the screen
	if(paused) GameX.DrawOutlinedText(screen_width/2-48,screen_height/2,"Paused",255,255,0);
	DrawCrosshair(); // draw a crosshair where the mouse cursor is
	DrawFade(); // fade screen black if the game's just starting up
}


void InitGameX (void)
{											// Initialize GameX
	GameX.Initialize ("MissileDemoX", VIDEO_16BIT|RUN_ALWAYSHIDECURSOR, 800, 600);
	screen_width = GameX.GetWidth();
	screen_height = GameX.GetHeight();
	ground_pos = ((int)(SKY_FRACTION*(float)screen_height));
}

void LoadImages (void)
{
	city_img.Load ("MissileDemoImages\\city.tif", true);
	crosshair_img.Load("MissileDemoImages\\crosshair.bmp");
	explode_img[0].Load("MissileDemoImages\\burst.bmp"); // red+yellow explosion
	explode_img[0].CopyTo(&explode_img[1]);
	explode_img[1].SwapColors(COLOR_SWAP_ROTATE120); // swap to be a green+cyan explosion
	explode_img[1].CopyTo(&explode_img[2]);
	explode_img[2].SwapColors(COLOR_SWAP_ROTATE120); // swap again to be a blue+magenta explosion
}

void PrepareColors (void)
{
	ColorX black = ColorX(0,0,0);
	ColorX earth = ColorX(160,100,20);
	groundColor = ColorX(black,black,earth,earth); // gradient of black on top to earth-color at bottom

	ColorX blue = ColorX(0,0,128);
	skyColor = ColorX(blue,blue,black,black); // gradient of blue on top to black at bottom

	ColorX white = ColorX(255,255,255,255);
	ColorX clear = ColorX(255,255,255,0);
	turretColor = ColorX(clear,clear,white,white); // gradient of clear at top to white at bottom
}

void InitCities (void)
{
	for (int i = 0 ; i < NUM_CITIES ; i++) {
		city[i].x = (int)((float)(i+1) / (float)(NUM_CITIES+1) * (float)screen_width);
		city[i].y = ground_pos;
		city[i].radius = (48 + (int)(32.0f*sinf((float)city[i].x/(float)screen_width*180.0f*DEGtoRAD)))*5/NUM_CITIES;
		city[i].alive = 1.0f;
		city[i].charged = 0.0f;

		// pick a color at random for this city:
		int r = GameX.GetRandomInt(96,255);
		int g = GameX.GetRandomInt(96,255);
		int b = GameX.GetRandomInt(96,255);
		int total = r + g + b; // normalize so that r+g+b = 512:
		city[i].color = ColorX(r*512/total, g*512/total, b*512/total);
		if(r >= g && r >= b)		city[i].explode_img_id = 0; // red-ish
		else if(g >= r && g >= b)	city[i].explode_img_id = 1; // green-ish
		else						city[i].explode_img_id = 2; // blue-ish
	}
}

void PositionMouse (void)
{
	// Get mouse coordinate values
	int mouseX = GameX.GetMouseX();
	int mouseY = GameX.GetMouseY();
	int newMouseX = mouseX;
	int newMouseY = mouseY;

	// Allow keyboard repositioning of the mouse here:
	if(GameX.IsKeyDown(KEY_LEFT) || GameX.IsKeyDown(KEY_A))  newMouseX -= 6;
	if(GameX.IsKeyDown(KEY_RIGHT) || GameX.IsKeyDown(KEY_D)) newMouseX += 6;
	if(GameX.IsKeyDown(KEY_UP) || GameX.IsKeyDown(KEY_W))    newMouseY -= 4;
	if(GameX.IsKeyDown(KEY_DOWN) || GameX.IsKeyDown(KEY_S))  newMouseY += 4;

	// Limit the mouse coordinate values
	if(newMouseY > ground_pos - 180)  newMouseY = ground_pos - 180;
	else if(newMouseY < 32)           newMouseY = 32;
	if(newMouseX > screen_width - 32) newMouseX = screen_width - 32;
	else if(newMouseX < 32)           newMouseX = 32;

	// If the new coordinates differ from the old, set mouse to new position:
	if(newMouseX != mouseX || newMouseY != mouseY)
		GameX.SetMousePosition(newMouseX, newMouseY);
}

void PositionTurrets (void)
{
	// Set each turret tip to be a certain distance along the path it'll shoot to:
	for (int i = 0 ; i < NUM_CITIES ; i++) {
		int xDiff = GameX.GetMouseX() - city[i].x;
		int yDiff = GameX.GetMouseY() - city[i].y;
		float dy = -sqrtf(fabsf(-yDiff*2*gravity));
		Vector2DF v = Vector2DF(-xDiff*gravity/dy, dy).Normalize()*(1.5f*city[i].radius*city[i].charged);
		city[i].turret_x = (int)(v.x);
		city[i].turret_y = (int)(v.y);
	}
}

void UpdateCities (void)
{
	bool gameGoing = false;
	for (int i = 0 ; i < NUM_CITIES ; i++) {
		// Slowly recharge cities that have fired a missile recently:
		if(city[i].charged < 1.0f) {
			city[i].charged += 0.005f;
			if(city[i].charged > 1.0f) city[i].charged = 1.0f;
		}
		// Advance destruction of city if it's being destroyed:
		if(city[i].alive < 1.0f && city[i].alive > -0.25f) {
			city[i].alive -= 0.01f;
		}
		// If any city is alive, we know it's not game-over yet
		if(city[i].alive > -0.25f)
			gameGoing = true;
	}
	if(!gameGoing) { // if game-over, for now just restart the gane
		InitCities();
		num_missiles = 0;
		fadingIn = 1.0f;
	}
}

void LaunchMissiles (void)
{
	// friendly missiles:
	if(GameX.GetMouseClick(MOUSE_LEFT) || GameX.GetKeyPress(KEY_SPACE) || GameX.GetKeyPress(KEY_X) // if regular fire just pressed
	|| GameX.IsMouseDown(MOUSE_RIGHT) || GameX.IsKeyDown(KEY_C) || GameX.IsAltDown()) {            // or if auto-fire is held
		if(num_missiles < MAX_MISSILES) {
			int nearCity = FindNearestCity(GameX.GetMouseX(),true); // find nearest city that can fire now
			if(nearCity != -1) {

				missile[num_missiles].x = city[nearCity].x + city[nearCity].turret_x;
				missile[num_missiles].y = city[nearCity].y + city[nearCity].turret_y;

				int xDiff = GameX.GetMouseX() - missile[num_missiles].x;
				int yDiff = GameX.GetMouseY() - missile[num_missiles].y;

				// set initial velocity such that it will reach its peak where the crosshair is:
				missile[num_missiles].dy = -sqrtf(fabsf(-yDiff*2*gravity));
				missile[num_missiles].dx = -xDiff*gravity/missile[num_missiles].dy;
				
				missile[num_missiles].origin = nearCity;
				missile[num_missiles].exploding = 0.0f;
				missile[num_missiles].max_explode_radius = GameX.GetRandomFloat(64.0f, 96.0f);
				missile[num_missiles].img_id = city[nearCity].explode_img_id;

				num_missiles++;

				// make the city that just fired have to wait/recharge a bit before firing again:
				city[nearCity].charged -= 72.0f/(float)city[nearCity].radius;
			}
		}
	}

	// enemy missiles:
	if((GameX.GetGlobalCounter()+1) % 160 == 0) { // add enemy missile every 160 60ths of a second,
		if(num_missiles < MAX_MISSILES) {
			missile[num_missiles].x = GameX.GetRandomFloat(32.0f,(float)screen_width-32.0f);
			missile[num_missiles].y = 0;

			missile[num_missiles].dy = GameX.GetRandomFloat(0.3f,1.0f);
			missile[num_missiles].dx = GameX.GetRandomFloat(0.0f,1.5f) * missile[num_missiles].dy;
			if(missile[num_missiles].x > (float)screen_width/2) missile[num_missiles].dx *= -1.0f;
			
			missile[num_missiles].origin = -1; // enemy
			missile[num_missiles].exploding = 0.0f;
			missile[num_missiles].max_explode_radius = GameX.GetRandomFloat(80.0f, 128.0f);
			missile[num_missiles].img_id = 0;

			num_missiles++;
		}
	}
}

void UpdateMissiles (void)
{
	for(int i = 0 ; i < num_missiles ; i++) {

		if(missile[i].origin != -1) // (for gameplay reasons, enemy missiles are immune to gravity)
			missile[i].dy += gravity;

		missile[i].x += missile[i].dx; // move missile by its velocity
		missile[i].y += missile[i].dy;

		if(missile[i].exploding == 0.0f) { // if missile is not exploding,

			bool explode = false;

			if(missile[i].origin != -1) {
				if(missile[i].dy >= 0.0f || missile[i].y < 0)
					explode = true; // explode if it's not an enemy missile and it's reached its peak
			} else {
				if(missile[i].y >= (float)ground_pos) {
					explode = true; // explode if it is an enemy missile and it's reached the ground

					int nearCity = FindNearestCity(missile[i].x,false);
					if(nearCity != -1) {
						int xDist = abs(missile[i].x - city[nearCity].x);
						if(xDist < (city[nearCity].radius+missile[i].max_explode_radius)/2) {
							// explosion hit ground and is too close, signal city to die:
							city[nearCity].alive -= 0.1f;

							// add an explosion of the city's color at its center:
							missile[num_missiles].x = (float)city[nearCity].x;
							missile[num_missiles].y = (float)city[nearCity].y;
							missile[num_missiles].dx = 0;
							missile[num_missiles].dy = -2.5f;
							missile[num_missiles].img_id = city[nearCity].explode_img_id;
							missile[num_missiles].origin = nearCity;
							missile[num_missiles].exploding = 3.0f;
							missile[num_missiles].max_explode_radius = 140.0f;
							num_missiles++;
						}
					}
				} else {
					// every once in a while (somewhat often, actually) split an enemy missile into two
					if(rand()%2048 == 0) {
						if(num_missiles < MAX_MISSILES) {
							memcpy(&missile[num_missiles],&missile[i],sizeof(Missile));
							missile[num_missiles].dx += GameX.GetRandomFloat(-0.1f,0.1f);
							missile[num_missiles].dx *= GameX.GetRandomFloat(0.5f,1.25f);
							missile[num_missiles].dy *= GameX.GetRandomFloat(0.75f,1.25f);
							num_missiles++;
						}
					}
				}
			}

			// check for collisions of this not-yet-exploded missile with other explosions
			if(!explode)
			for(int j = 0 ; j < num_missiles ; j++) {
				if(missile[j].exploding > 0.0f && missile[j].exploding < missile[j].max_explode_radius*0.875f) {
					float radiusSq = missile[j].exploding * missile[j].exploding;
					float xDiff = missile[i].x-missile[j].x;
					float yDiff = missile[i].y-missile[j].y;
					float distSq = xDiff*xDiff+yDiff*yDiff;
					if(distSq < radiusSq) {
						explode = true;
						break;
					}
				}
			}


			if(explode) {
				missile[i].angle = GameX.GetRandomFloat(0.0f, 360.0f); // pick a random angle to explode at
				missile[i].exploding = 1.5f; // and start the explosion
			}

		} else { // if missile is exploding,

			missile[i].dx *= 0.925f; // heavily dampen the velocities of explosions
			missile[i].dy *= 0.875f;

			// advance the explosion outward:
			missile[i].exploding += 0.01f + missile[i].max_explode_radius/5/missile[i].exploding;

			// if the explosion has advanced so far that it's over,
			if(missile[i].exploding > missile[i].max_explode_radius)  // delete exploding missile by putting
				memcpy (&missile[i], &missile[--num_missiles], sizeof(Missile)); // the last one in its place
		}
	}
}

void UpdateFade (void)
{
	if(fadingIn > 0.0f) {
		fadingIn -= 0.005f;
		if(fadingIn < 0.0f) fadingIn = 0.0f;
	}
}

void UpdatePause (void)
{
	if(GameX.GetKeyPress(KEY_P))
		paused = !paused;
}

void DrawFade (void)
{
	if(fadingIn > 0.0f) {
		GameX.BlurScreen(fadingIn);
		GameX.FillScreen(ColorX(0,0,0,fadingIn*255));
		if(!paused) // also draw some very simple instructions while we're just starting up:
			GameX.DrawOutlinedText(screen_width/2-112,screen_height/2,"Click to shoot down enemy missiles!",255,0,0);
	}
}

void DrawCrosshair (void)
{
	GameX.SetDrawMode(DRAW_ADD  | DRAWOP_CENTERED);
	GameX.DrawImage(&crosshair_img, GameX.GetMouseX(), GameX.GetMouseY(), -GameX.GetGlobalCounter(), GameX.GetOscillatingValue(1.25f,1.75f,1.0f));
}

void DrawGround (void)
{
	GameX.DrawRect(groundColor, 0, ground_pos, screen_width, screen_height);
}

void DrawLiveCities (void)
{
	DrawCities(true);
}

void DrawDyingCities (void)
{
	DrawCities(false);
}

void DrawCities (bool live)
{
	// Find the city that will fire next so we can visually highlight it:
	int nearCity = FindNearestCity(GameX.GetMouseX(),true);

	for (int i = 0 ; i < NUM_CITIES ; i++) { // for each city...
		if(city[i].alive == 1.0f && live || city[i].alive < 1.0f && !live) {

			// Draw the city's turret:
			if(city[i].alive > 0.0f) {
				int under;
				if(city[i].alive != 1.0f) under = (int)((1.0f-city[i].alive) * (screen_height*GROUND_FRACTION));
				else under = 0;
				GameX.SetDrawScale(4.0f);
				GameX.DrawLine(turretColor, city[i].x, city[i].y+under, city[i].x+city[i].turret_x, city[i].y+under+city[i].turret_y);
			}
			
			// Draw the city itself:
			GameX.SetDrawMode(DRAW_PLAIN | DRAWOP_CENTERED | ((i == nearCity) ? DRAWOP_BRIGHT : 0));
			GameX.SetDrawShading(city[i].color);
			int yCenter = city[i].y-(int)((float)city[i].radius*city[i].alive);
			GameX.DrawImage(&city_img, city[i].x, yCenter, 0, (float)city[i].radius/64.0f);
		}
	}
}

void DrawMissiles (void)
{
	for(int i = 0 ; i < num_missiles ; i++) { // for each non-exploding missile...
		int o = missile[i].origin;
		if(missile[i].exploding == 0.0f) {
			GameX.SetDrawScale(3.0f);
			if(o == -1) { // enemy missile, all drawn the same length
				Vector2DF v = Vector2DF(missile[i].dx, missile[i].dy).Normalize()*20.0f;
				GameX.DrawLine(ColorX(255,0,0), missile[i].x-v.x, missile[i].y-v.y, missile[i].x+v.x, missile[i].y+v.y);
			} else { // otherwise, friendly missile, drawn so length varies with speed
				GameX.DrawLine(city[o].color, missile[i].x-10*missile[i].dx, missile[i].y-10*missile[i].dy, missile[i].x, missile[i].y);
			}
		}
	}
}

void DrawExplosions (void)
{
	for(int i = 0 ; i < num_missiles ; i++) { // for each exploding missile...
		int o = missile[i].origin;
		if(missile[i].exploding != 0.0f) {
			GameX.SetDrawScale(missile[i].exploding/32.0f);
			GameX.SetDrawAngle(missile[i].angle + 90.0f*missile[i].dx+(float)GameX.GetGlobalCounter()*0.1f);
			GameX.SetDrawMode(DRAW_ADD | DRAWOP_BRIGHT | DRAWOP_CENTERED);

			// Set the shading color to match the exploding missile's origin:
			float brightness = (missile[i].max_explode_radius-missile[i].exploding)*2.0f/missile[i].max_explode_radius;
			ColorX c = (o == -1) ? ColorX(255,192,128) : city[o].color; // enemy missile=orange-red, otherwise shade like city
			GameX.SetDrawShading(c * brightness);

			// In addition to shading it, also choose an image with colors loosely matching the color
			// (red-ish, green-ish, or blue-ish), and draw it additively at the explosion center:
			GameX.DrawImage(&explode_img[missile[i].img_id], missile[i].x, missile[i].y);
		}
	}
}

// Returns the array index of the city with the least distance from x
int FindNearestCity (int x, bool must_be_charged)
{
	int minDist = 99999;
	int minCity = -1;
	for (int i = 0 ; i < NUM_CITIES ; i++) {
		if(city[i].alive <= 0.0f || (must_be_charged && city[i].charged < 1.0f)) continue;
		int dist = abs(x - city[i].x);
		if(dist < minDist) {
			minDist = dist;
			minCity = i;
		}
	}
	return minCity;
}
