//
// GameX
// Puzzle Demo - Main Program
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#define GAMEX_MAIN
#pragma comment (lib, "gamex.lib")
#include "gamex.hpp"

#include "puzzledemox-text.hpp"
#include "puzzledemox-resourcepool.hpp"
											//*** Global Variables

#define KEY_DELAY_INTERVAL 14 // delay between held keypresses being re-recognized
int keyDelay = 0; // leftover delay until keys being held are counted as if it were pressed again
int controlling = true;

const int scoreLeft = 630, scoreTop = 100, scoreRight = 970, scoreBottom = 270;
int score = 0;

const int boardLeft = 120, boardTop = 20, boardRight = 520, boardBottom = 836;
#define BOARD_COLS 6
#define BOARD_ROWS 13

struct Piece {						//*** Particle structure
	int row, col, color;			// Row, column, and color of piece on board
	bool connected;					// false if piece is still falling
	int destroyed;					// 0 if alive, 1-63 if dying, 64 if dead
};

#define MAX_PIECES (BOARD_COLS*BOARD_ROWS+16)
Piece piece [MAX_PIECES];
int num_pieces = 0;

#define MATCH_NUMBER 3 // number of pieces that must match at once
#define SEL_PIECE_NUM 3 // pieces that fall at once
int selectedPieces [SEL_PIECE_NUM]; // currently being-controlled piece indices
bool addPiece = true; // if true, we add a new piece as soon as possible
int rowHeight [BOARD_ROWS];

int variety = 3; // difficulty level, should be between 2 and 8, can be increased mid-game
int fallDelay = 38; // cycles between piece(s) falling an entire row, can be decreased mid-game
#define DELAY_MIN 8
int fallDelayDiv = 1; // temporary delay divisor
int fallCounter = 0;
bool gameOver = false;
											//*** Custom Function Headers
int checkMatches(void);
void drawBackground(void);
void drawBoard(void);
void drawPieces(void);
void drawPiece(int n);
void drawZapLine(int x1, int y1, int x2, int y2, float width, int mode, int slowness);
int addPieceAt(int row, int col); // returns piece's index #
bool isOccupied(int row, int col); // returns true if location is occupied or illegal, false otherwise
int findPieceAt(int row, int col); // returns index # of piece at location
void preDestroy(int n);
void resetGame(void);
void handlePieceDestruction(void);
void handleFallingAndMatches(void);
void handlePieceAddition(void);
void handleControls(void);

// Required GameX initialize function (called when program starts)
// this function should initialize GameX and prepare anything else we need
void GameInit(void)
{											// Initialize GameX itself (required)
	GameX.Initialize("PuzzleDemo",VIDEO_16BIT, 1070, 856); // (very weird game size)

	tp.Initialize("PuzzleDemoImages\\text.gif");
	rp.Initialize();

	GameX.PlayMusic(&rp.carnival_song); // start playing the music immediately

	resetGame();
}

// Required GameX run function (called repeatedly while game is running)
// this function should update the game state but should NOT draw that state
void GameRun(void)
{
	if(gameOver) {
		if(GameX.AnyKeyPress()) resetGame();
		else return;
	}

	handlePieceDestruction();
	handleFallingAndMatches();
	handlePieceAddition();
	handleControls();
}

// Required GameX draw function (called repeatedly while game is running)
// this function should draw the game state and do as little else as possible
void GameDraw(void)
{
	drawBackground();
	drawBoard();
	drawPieces();
	if(gameOver) {
		GameX.FillScreen(ColorX(0,0,0,128));
		tp.PrintText(DRAW_GHOST|DRAWOP_BRIGHT, 36.0f, "GAME OVER", 210,440, ColorX(255,GameX.GetRandomInt(0,255),0));
	}
}

void drawBackground(void)
{
	GameX.SetDrawAngle(-(float)GameX.GetGlobalCounter()/100.0f); // make the background *slowly* spin clockwise
	GameX.DrawTiled(&rp.stars_img);
}

void drawBoard(void)
{
	GameX.DrawRect(ColorX(0,0,255,0, 0,0,255,0, 96,128,255,128, 96,128,255,128), boardLeft,boardTop,boardRight,boardBottom);
	drawZapLine(boardLeft,boardTop,boardLeft,boardBottom, 10.0f, DRAW_BURN, 6);
	drawZapLine(boardRight,boardTop,boardRight-3,boardBottom, 10.0f, DRAW_BURN, 7);
	drawZapLine(boardLeft,boardBottom,boardRight,boardBottom, 10.0f, DRAW_BURN, 8);

	GameX.DrawRect(ColorX(255,0,0,0, 0,128,128,128, 128,128,0,128, 0,0,255,255), scoreLeft,scoreTop,scoreRight,scoreBottom);
	drawZapLine(scoreLeft,scoreTop,scoreRight,scoreTop, 10.0f, DRAW_BURN, 6);
	drawZapLine(scoreLeft,scoreTop,scoreLeft,scoreBottom, 10.0f, DRAW_BURN, 7);
	drawZapLine(scoreRight,scoreTop,scoreRight-3,scoreBottom, 10.0f, DRAW_BURN, 8);
	drawZapLine(scoreLeft,scoreBottom,scoreRight,scoreBottom, 10.0f, DRAW_BURN, 9);

	char str [64];
	sprintf(str, "Score: %d", score);
	tp.PrintText(DRAW_ADD, 24.0f, str, scoreLeft + 75, scoreTop + 75);
}

void drawPieces(void)
{
	for(int i = 0 ; i < num_pieces ; i++) {
		if(piece[i].destroyed)
			drawPiece(i);
	}
	for(int j = 0 ; j < num_pieces ; j++) {
		if(!piece[j].destroyed)
			drawPiece(j);
	}
}

void drawPiece(int n)
{
	int left = (piece[n].color < 4) ? 0 : 128;
	int top = (piece[n].color % 4) * 64;
	GameX.SetDrawPart(left,top,left+64,top+64);
	bool selected = false;
	for(int i = 0 ; i < SEL_PIECE_NUM ; i++)
		if(n == selectedPieces[i])
			selected = true;
	DrawFlags mode = DRAWOP_CENTERED;
	if(piece[n].destroyed) { // fading out: (if being destroyed)
		mode |= DRAW_ADD;
		GameX.SetDrawShading(ColorX(255,255,255,255-4*piece[n].destroyed));
		GameX.SetDrawScale(1.0f + sqrtf(9.0f * (float)piece[n].destroyed/64.0f));
	} else if(selected && !piece[n].connected) { // glowing: (if being controlled)
		mode |= DRAWOP_BRIGHT;
		int bright = (int)GameX.GetOscillatingValue(128.0f,255.0f,4.0f);
		GameX.SetDrawShading(ColorX(bright,bright,bright));
	}
	GameX.SetDrawMode(mode);
	GameX.DrawImage(&rp.aliens_img, boardLeft + 64*piece[n].col + 6+32, boardBottom - 64*piece[n].row - 64+32);
}

void drawZapLine(int x1, int y1, int x2, int y2, float width, int mode, int slowness)
{
	int frame = (GameX.GetGlobalCounter()/slowness)%4;
	GameX.SetDrawPart(frame*32,0,(frame+1)*32,128);
	GameX.SetDrawScale(width);
	GameX.SetDrawMode(mode);
	GameX.DrawLineImage(&rp.zap_img, x1, y1, x2, y2, true);
}

int addPieceAt(int row, int col)
{
	int index = findPieceAt(row,col);
	if(index != -1) return -1;

	piece[num_pieces].color = rand()%variety;
	piece[num_pieces].row = row;
	piece[num_pieces].col = col;
	piece[num_pieces].connected = false;
	piece[num_pieces].destroyed = 0;

	return num_pieces++; // returns num_pieces and THEN increments it for next time
}

bool isOccupied(int row, int col)
{
	if(row < 0 || row >  BOARD_ROWS) return true;
	if(col < 0 || col >= BOARD_COLS) return true;
	return(row <= rowHeight[col]);
}

int findPieceAt(int row, int col)
{
	int n;
	for(n = 0; n < num_pieces; n++)
		if(piece[n].row == row && piece[n].col == col && piece[n].destroyed <= 1 && piece[n].connected)
			return n;
	return -1;
}

int checkMatches(void)
{
	int total = 0;
	int stack[MAX_PIECES];
	int aMax, bMax;
	for(int dir = 0 ; dir < 2 ; dir++) {
		switch(dir) {
			case 0: aMax = BOARD_ROWS, bMax = BOARD_COLS; break;
			case 1: aMax = BOARD_COLS, bMax = BOARD_ROWS; break;
		}
		for(int a = 0 ; a <= aMax ; a++) {
			int num = 0, run = 0, lastColor = -1;
			for(int b = 0 ; b <= bMax ; b++) {
				int index = (dir == 0) ? findPieceAt(a, b) : findPieceAt(b, a);
				int color = (index == -1) ? (-1) : piece[index].color;
				if(lastColor != color) {
					if(run >= MATCH_NUMBER) {
						total += num;
						while(num) preDestroy(stack[--num]);
					} else num = 0;
					lastColor = color;
					run = 0;
				}
				if(index != -1) stack[num++] = index;
				run++;
			}
		}
	}

	for(int n = 0 ; n < num_pieces ; n++)
		if(piece[n].row > rowHeight[piece[n].col] && !piece[n].destroyed)
			piece[n].connected = false;

	{
		addPiece = true;
		for(int n = 0 ; n < num_pieces ; n++)
			if(!piece[n].connected)
				addPiece = false;
		controlling = addPiece;
	}

	return total;
}

void preDestroy(int n)
{
	if(!piece[n].destroyed) {
		piece[n].destroyed = 1;
		rowHeight[piece[n].col] = min(piece[n].row, rowHeight[piece[n].col]);
	}
}

void resetGame(void)
{
	int i;
	for(i = 0 ; i < BOARD_ROWS ; i++)
		rowHeight[i] = 0;
	for(i = 0 ; i < SEL_PIECE_NUM ; i++)
		selectedPieces[i] = -1;

	num_pieces = 0;
	addPiece = true;
	score = 0;
	gameOver = false;
}

void handlePieceDestruction(void)
{
	for(int n = 0 ; n < num_pieces ; n++) {
		if(piece[n].destroyed) {
			piece[n].destroyed++;
			if(piece[n].destroyed > 63) {
				memcpy(&piece[n], &piece[--num_pieces], sizeof(Piece)); // replace with last piece
				for(int i = 0 ; i < SEL_PIECE_NUM ; i++)
					if(selectedPieces[i] == num_pieces)
						selectedPieces[i] = n;
				n = 0;
			}
		}
	}
}

void handleFallingAndMatches(void)
{
	if(++fallCounter % (max(DELAY_MIN,fallDelay/fallDelayDiv)) == 0) {
		bool landed = false;
		for(int n = 0 ; n < num_pieces ; n++) {
			if(!piece[n].connected) {
				if(isOccupied(piece[n].row, piece[n].col)) { // piece just landed
					landed = true;
					piece[n].connected = true;
					piece[n].row = rowHeight[piece[n].col]; // in case of positioning error
					rowHeight[piece[n].col]++;
					n = -1; // restart landing search in case of simultaneous stack landing
				}
			}
		}

		if(landed) {
			int matches = checkMatches();

			int lastScore = score;
			score += 10*(variety-1)*(int)(powf((float)matches, 1.5f)) + 1;
			if(lastScore < 500   && score >= 500
			|| lastScore < 4000  && score >= 4000
			|| lastScore < 10000 && score >= 10000
			|| lastScore < 18000 && score >= 18000
			|| lastScore < 28000 && score >= 28000) {
				variety++;
				fallDelay -= 6;
			}

			if(matches)
				GameX.PlaySound(&rp.poof_snd, PLAY_REWIND, 1.0f, 0.0f, (float)matches / (float)MATCH_NUMBER);
			else
				GameX.PlaySound(&rp.stick_snd, PLAY_REWIND, 1.0f, 0.0f, (controlling?0.75f:0.625f));
		}
		if(!landed || !controlling) {
			for(int n = 0 ; n < num_pieces ; n++)
				if(!piece[n].connected && !piece[n].destroyed)
					if(!isOccupied(piece[n].row, piece[n].col))
						piece[n].row--;
		}

	}
}

void handlePieceAddition(void)
{
	if(addPiece) {
		addPiece = false;
		if(num_pieces+1<MAX_PIECES) {
			for(int i = 0 ; i < SEL_PIECE_NUM ; i++) {
				selectedPieces[i] = addPieceAt(BOARD_ROWS-(SEL_PIECE_NUM-i),3);
				if(selectedPieces[i] == -1) {
					gameOver = true;
					GameX.PlaySound(&rp.gameover_snd);
					break;
				}
			}
		} else {
			GameX.ErrorDialog("Piece-adding failed -- too many pieces.");
		}
	}
}

void handleControls(void)
{
	if(!controlling || GameX.IsKeyDown(KEY_SPACE))
		fallDelayDiv = 8; // speed up game
	else
		fallDelayDiv = 1;

	if(controlling) {
		int ySwitch = 0;
		if((keyDelay == 0 && GameX.IsKeyDown(KEY_UP))   || GameX.GetKeyPress(KEY_UP))   ySwitch++;
		if((keyDelay == 0 && GameX.IsKeyDown(KEY_DOWN)) || GameX.GetKeyPress(KEY_DOWN)) ySwitch--;

		if(ySwitch) {
			GameX.PlaySound(&rp.spin_snd, PLAY_REWIND, 0.4f);
			int rows [SEL_PIECE_NUM];
			for(int i = 0 ; i < SEL_PIECE_NUM ; i++)
				rows[i] = piece[selectedPieces[i]].row;
			for(int j = 0 ; j < SEL_PIECE_NUM ; j++)
				piece[selectedPieces[j]].row = rows[(j+ySwitch+SEL_PIECE_NUM)%SEL_PIECE_NUM];
		}

		int xPlus = 0;
		if((keyDelay == 0 && GameX.IsKeyDown(KEY_LEFT))  || GameX.GetKeyPress(KEY_LEFT))  xPlus--;
		if((keyDelay == 0 && GameX.IsKeyDown(KEY_RIGHT)) || GameX.GetKeyPress(KEY_RIGHT)) xPlus++;


		if(xPlus) {
			for(int i = 0 ; i < SEL_PIECE_NUM ; i++) {
				if(isOccupied(piece[selectedPieces[i]].row+1, xPlus + piece[selectedPieces[i]].col)) {
					xPlus = 0;
				}
			}
			for(int j = 0 ; j < SEL_PIECE_NUM ; j++) {
				piece[selectedPieces[j]].col += xPlus;
			}
		}

		if(xPlus || ySwitch)
			keyDelay = KEY_DELAY_INTERVAL;
	}

	if(keyDelay > 0) keyDelay--;
}
