#ifndef SPRITE_DEFINED
	#define SPRITE_DEFINED
	#include "gamex.hpp"
	#include "resourcepool.hpp"
	#include "text.hpp"

	#define MAX_TRAIL_LENGTH ((int)512)

	enum spriteType {
		SPRITE_MAIN_CHAR, SPRITE_ENEMY, SPRITE_FIREBALL, SPRITE_ICEBALL, SPRITE_STAR
	};

	class Sprite
	{
		public:
			// Sprite Constructor: creates this sprite using the following parameters:
			// -type should be a spriteType (SPRITE_MAIN_CHAR, SPRITE_FIREBALL, etc.)
			// -trail is the trail length, 0 for no trail, MAX_TRAIL_LENGTH for longest trail
			// -initialPoint is the initial positions of the sprite
			// NOTE: if not using 3D positioning, use 0 for z: sprite = new Sprite(type,trail,Vector3DF(x,y,0));
			Sprite(int type, int trail_length, Vector3DF initialPoint);
			~Sprite(void);

			void update(void); // call once per cycle for active sprites
			void draw(void); // call once per frame to tell sprite to draw itself

			// accelerate(): adds the given vector to this sprite's velocity
			void accelerate(Vector3DF accelerationVector);

			// attractTo(): attracts this sprite linearly to a location
			// -attractionPoint is the point of attraction
			// -multiple is a factor for how quickly it should move toward the attraction point
			void attractTo(Vector3DF attractionPoint, float multiple);

			// gravityAttractTo(): attracts this sprite gravitationally to a location
			// -attractionPoint is the point of attraction
			// -multiple is a factor for how quickly it should move toward the attraction point
			void gravityAttractTo(Vector3DF attractionPoint, float multiple);

			// enforceMaxDistance(): moves this sprite closer to a location if it's too far away
			// -centerPoint is the point of attraction, centerPoint.z should be 0 if only using 2D positioning
			// -maxDist is a (pixel) radius beyond which this sprite will be moved closer to the center
			// -pullMultiple is a factor for how much to multiply the distance if beyond the maxDist, 1 means maxDist has no effect, 0.9 means it moves 10% closer to the attraction point each instant it's outside the maxDist
			void enforceMaxDistance(Vector3DF centerPoint, float maxDist, float pullMultiple=0.9f);

			// multiplyVelocity(): multiplies this sprite's velocity vector by the given multiple m
			void multiplyVelocity(float m);
			
			// collideCheck(): checks for and handles collision with this and another sprite
			void collideCheck(Sprite* s);

			// moveTo(): instantly moves this sprite to a new location
			void moveTo(Vector3DF newLocation);

			int type; // type of this sprite -- a spriteType (SPRITE_MAIN_CHAR, SPRITE_FIREBALL, etc.)
			int trailLength; // length of visible trail behind sprite, 0 for no trail
			Vector3DF position; // current position of sprite. +x is right, +y is down, +z is in
			Vector3DF velocity; // current velocity of sprite

			int current; // index of x,y arrays that gives most recent trail position
			Vector3DF trailPosition[MAX_TRAIL_LENGTH]; // recent positions
			Vector3DF trailVelocity[MAX_TRAIL_LENGTH]; // velocities moving recent positions after the fact

			int curr_img; // current image number, for sprites consisting of multiple possible images
			int damage; // damage amount, for sprites that can take damage
	};
#endif