// GameX (Release 2 - Vectors / Matricies / CameraX)
// Vector Class Header
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
//
//
// Vector Operations Implemented:
// -- operations that do not modify the vector:
//		+, -, *, /      Term-by-term operations on vectors, * and / work with a scalar also
//		Dot				Dot product vector with another vector
//		Cross			Cross product of vector with another vector
//		Direction		Gives a unit-length vector with the same direction as this one
//		Length			Length of a vector
//		LengthSq		Length^2 of a vector
//		Dist			Distance from vector to another vector
//		DistSq			Distance^2 from vector to another vector
// -- operations that modify the vector being operated on:
//      =, +=, -=, *=, /=	Term-by-term modifications, for example: a += b sets a to be the result of a + b
//		CrossWith			Sets vector to be the cross product of itself with another vector (3D vectors only)
//		Normalize			Sets vector to be its Direction vector
//

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <memory.h>
#include <math.h>

#ifndef VECTOR_DEF
	#define VECTOR_DEF

											// **** Switches	
	#define VECTOR_DEBUG					// Enable Debugging Output
	#define VECTOR_INITIALIZE				// Initializes vectors	
																// Forward Referencing
	class Vector2DI;
	class Vector2DF;
	class Vector3DI;
	class Vector3DF;
	class Vector4DF;
	class MatrixF;

	// Vector2DI Declaration

	#define VNAME		2DI
	#define VTYPE		int

	class Vector2DI {
	public:
		VTYPE x, y;

		// Constructors/Destructors
		inline Vector2DI();							
		inline ~Vector2DI();			
		inline Vector2DI (VTYPE xa, VTYPE ya);
		inline Vector2DI (const Vector2DI &op);				
		inline Vector2DI (const Vector2DF &op);				
		inline Vector2DI (const Vector3DI &op);				
		inline Vector2DI (const Vector3DF &op);				
		inline Vector2DI (const Vector4DF &op);

		// Member Functions
		inline Vector2DI &operator= (const Vector2DI &op);

		inline Vector2DI &operator+= (const Vector2DI &op);
		inline Vector2DI  operator+  (const Vector2DI &op) const {Vector2DI v = Vector2DI(*this); v += op; return v;}

		inline Vector2DI &operator-= (const Vector2DI &op);
		inline Vector2DI  operator-  (const Vector2DI &op) const {Vector2DI v = Vector2DI(*this); v -= op; return v;}

		inline Vector2DI &operator*= (int op);
		inline Vector2DI &operator*= (float op);
		inline Vector2DI &operator*= (const Vector2DI &op);
		inline Vector2DI  operator*  (int op) const              {Vector2DI v = Vector2DI(*this); v *= op; return v;}
		inline Vector2DI  operator*  (float op) const            {Vector2DI v = Vector2DI(*this); v *= op; return v;}
		inline Vector2DI  operator*  (const Vector2DI &op) const {Vector2DI v = Vector2DI(*this); v *= op; return v;}

		inline Vector2DI &operator/= (int op);
		inline Vector2DI &operator/= (float op);
		inline Vector2DI &operator/= (const Vector2DI &op);
		inline Vector2DI  operator/  (int op) const              {Vector2DI v = Vector2DI(*this); v /= op; return v;}
		inline Vector2DI  operator/  (float op) const            {Vector2DI v = Vector2DI(*this); v /= op; return v;}
		inline Vector2DI  operator/  (const Vector2DI &op) const {Vector2DI v = Vector2DI(*this); v /= op; return v;}

		// Note: The "2D cross product" is the equivalent of rotating 90 degrees CCW and taking the dot product
		inline float Cross (const Vector2DI &op) const;
		inline float Dot (const Vector2DI &op) const;
		inline float Dist (const Vector2DI &op) const;
		inline float DistSq (const Vector2DI &op) const;
		
		inline Vector2DI &Normalize (void);
		inline Vector2DI  Direction (void) const {Vector2DI v = Vector2DI(*this); v.Normalize(); return v;}

		inline float Length (void) const;
		inline float LengthSq (void) const;

		inline VTYPE *Data (void);
	};

	inline Vector2DI operator * (const int op, const Vector2DI &v) {return v * op;}
	inline Vector2DI operator * (const float op, const Vector2DI &v) {return v * op;}
	inline Vector2DI operator / (const int op, const Vector2DI &v) {return Vector2DI((VTYPE)op/v.x, (VTYPE)op/v.y);}
	inline Vector2DI operator / (const float op, const Vector2DI &v) {return Vector2DI((VTYPE)op/v.x, (VTYPE)op/v.y);}

	#undef VNAME
	#undef VTYPE

	// Vector2DF Declarations

	#define VNAME		2DF
	#define VTYPE		float

	class Vector2DF {
	public:
		VTYPE x, y;

		// Constructors/Destructors
		inline Vector2DF ();
		inline ~Vector2DF ();
		inline Vector2DF (VTYPE xa, VTYPE ya);
		inline Vector2DF (const Vector2DI &op);
		inline Vector2DF (const Vector2DF &op);
		inline Vector2DF (const Vector3DI &op);
		inline Vector2DF (const Vector3DF &op);
		inline Vector2DF (const Vector4DF &op);

		// Member Functions
		inline Vector2DF &operator= (const Vector2DF &op);
		
		inline Vector2DF &operator+= (const Vector2DF &op);
		inline Vector2DF  operator+  (const Vector2DF &op) const {Vector2DF v = Vector2DF(*this); v += op; return v;}

		inline Vector2DF &operator-= (const Vector2DF &op);
		inline Vector2DF  operator-  (const Vector2DF &op) const {Vector2DF v = Vector2DF(*this); v -= op; return v;}

		inline Vector2DF &operator*= (int op);
		inline Vector2DF &operator*= (float op);
		inline Vector2DF &operator*= (const Vector2DF &op);
		inline Vector2DF  operator*  (int op) const              {Vector2DF v = Vector2DF(*this); v *= op; return v;}
		inline Vector2DF  operator*  (float op) const            {Vector2DF v = Vector2DF(*this); v *= op; return v;}
		inline Vector2DF  operator*  (const Vector2DF &op) const {Vector2DF v = Vector2DF(*this); v *= op; return v;}

		inline Vector2DF &operator/= (int op);
		inline Vector2DF &operator/= (float op);
		inline Vector2DF &operator/= (const Vector2DF &op);
		inline Vector2DF  operator/  (int op) const              {Vector2DF v = Vector2DF(*this); v /= op; return v;}
		inline Vector2DF  operator/  (float op) const            {Vector2DF v = Vector2DF(*this); v /= op; return v;}
		inline Vector2DF  operator/  (const Vector2DF &op) const {Vector2DF v = Vector2DF(*this); v /= op; return v;}

		// Note: The "2D cross product" is the equivalent of rotating 90 degrees CCW and taking the dot product
		inline float Cross (const Vector2DF &op) const;
		inline float Dot (const Vector2DF &op) const;
		inline float Dist (const Vector2DF &op) const;
		inline float DistSq (const Vector2DF &op) const;

		inline Vector2DF &Normalize (void);
		inline Vector2DF  Direction (void) const {Vector2DF v = Vector2DF(*this); v.Normalize(); return v;}

		inline float Length (void) const;
		inline float LengthSq (void) const;

		inline VTYPE *Data (void);
	};

	inline Vector2DF operator * (const int op, const Vector2DF &v) {return v * op;}
	inline Vector2DF operator * (const float op, const Vector2DF &v) {return v * op;}
	inline Vector2DF operator / (const int op, const Vector2DF &v) {return Vector2DF((VTYPE)op/v.x, (VTYPE)op/v.y);}
	inline Vector2DF operator / (const float op, const Vector2DF &v) {return Vector2DF((VTYPE)op/v.x, (VTYPE)op/v.y);}

	#undef VNAME
	#undef VTYPE

	// Vector3DI Declaration

	#define VNAME		3DI
	#define VTYPE		int

	class Vector3DI {
	public:
		VTYPE x, y, z;
	
		// Constructors/Destructors
		inline Vector3DI();
		inline ~Vector3DI();
		inline Vector3DI (VTYPE xa, VTYPE ya, VTYPE za);
		inline Vector3DI (const Vector2DI &op);
		inline Vector3DI (const Vector2DF &op);
		inline Vector3DI (const Vector3DI &op);
		inline Vector3DI (const Vector3DF &op);
		inline Vector3DI (const Vector4DF &op);

		// Member Functions
		inline Vector3DI &operator= (const Vector3DI &op);

		inline Vector3DI &operator+= (const Vector3DI &op);
		inline Vector3DI  operator+  (const Vector3DI &op) const {Vector3DI v = Vector3DI(*this); v += op; return v;}

		inline Vector3DI &operator-= (const Vector3DI &op);
		inline Vector3DI  operator-  (const Vector3DI &op) const {Vector3DI v = Vector3DI(*this); v -= op; return v;}

		inline Vector3DI &operator*= (int op);
		inline Vector3DI &operator*= (float op);
		inline Vector3DI &operator*= (const Vector3DI &op);
		inline Vector3DI  operator*  (int op) const              {Vector3DI v = Vector3DI(*this); v *= op; return v;}
		inline Vector3DI  operator*  (float op) const            {Vector3DI v = Vector3DI(*this); v *= op; return v;}
		inline Vector3DI  operator*  (const Vector3DI &op) const {Vector3DI v = Vector3DI(*this); v *= op; return v;}

		inline Vector3DI &operator/= (int op);
		inline Vector3DI &operator/= (float op);
		inline Vector3DI &operator/= (const Vector3DI &op);
		inline Vector3DI  operator/  (int op) const              {Vector3DI v = Vector3DI(*this); v /= op; return v;}
		inline Vector3DI  operator/  (float op) const            {Vector3DI v = Vector3DI(*this); v /= op; return v;}
		inline Vector3DI  operator/  (const Vector3DI &op) const {Vector3DI v = Vector3DI(*this); v /= op; return v;}

		inline Vector3DI &CrossWith (const Vector3DI &op);
		inline Vector3DI  Cross (const Vector3DI &op) const {Vector3DI v = Vector3DI(*this); v.CrossWith(op); return v;}

		inline float Dot (const Vector3DI &op) const;
		inline float Dist (const Vector3DI &op) const;
		inline float DistSq (const Vector3DI &op) const;

		inline Vector3DI &Normalize (void);
		inline Vector3DI  Direction (void) const {Vector3DI v = Vector3DI(*this); v.Normalize(); return v;}

		inline float Length (void) const;
		inline float LengthSq (void) const;

		inline VTYPE *Data (void);
	};
	
	inline Vector3DI operator * (const int op, const Vector3DI &v) {return v * op;}
	inline Vector3DI operator * (const float op, const Vector3DI &v) {return v * op;}
	inline Vector3DI operator / (const int op, const Vector3DI &v) {return Vector3DI((VTYPE)op/v.x, (VTYPE)op/v.y, (VTYPE)op/v.z);}
	inline Vector3DI operator / (const float op, const Vector3DI &v) {return Vector3DI((VTYPE)op/v.x, (VTYPE)op/v.y, (VTYPE)op/v.z);}

	#undef VNAME
	#undef VTYPE

	// Vector3DF Declarations

	#define VNAME		3DF
	#define VTYPE		float

	class Vector3DF {
	public:
		VTYPE x, y, z;
	
		// Constructors/Destructors
		inline Vector3DF();
		inline ~Vector3DF();
		inline Vector3DF (VTYPE xa, VTYPE ya, VTYPE za);
		inline Vector3DF (const Vector2DI &op);
		inline Vector3DF (const Vector2DF &op);
		inline Vector3DF (const Vector3DI &op);
		inline Vector3DF (const Vector3DF &op);
		inline Vector3DF (const Vector4DF &op);

		// Set Functions
		inline Vector3DF &Set (float xa, float ya, float za);
		
		// Member Functions
		inline Vector3DF &operator= (const Vector3DF &op);

		inline Vector3DF &operator+= (const Vector3DF &op);
		inline Vector3DF  operator+  (const Vector3DF &op) const {Vector3DF v = Vector3DF(*this); v += op; return v;}

		inline Vector3DF &operator-= (const Vector3DF &op);
		inline Vector3DF  operator-  (const Vector3DF &op) const {Vector3DF v = Vector3DF(*this); v -= op; return v;}

		inline Vector3DF &operator*= (int op);
		inline Vector3DF &operator*= (float op);
		inline Vector3DF &operator*= (const Vector3DF &op);
		Vector3DF &operator*= (MatrixF &op);				// see vector.cpp

		inline Vector3DF  operator*  (int op) const              {Vector3DF v = Vector3DF(*this); v *= op; return v;}
		inline Vector3DF  operator*  (float op) const            {Vector3DF v = Vector3DF(*this); v *= op; return v;}
		inline Vector3DF  operator*  (const Vector3DF &op) const {Vector3DF v = Vector3DF(*this); v *= op; return v;}
		Vector3DF  operator*  (MatrixF &op) const                {Vector3DF v = Vector3DF(*this); v *= op; return v;}

		inline Vector3DF &operator/= (int op);
		inline Vector3DF &operator/= (float op);
		inline Vector3DF &operator/= (const Vector3DF &op);

		inline Vector3DF  operator/  (int op) const              {Vector3DF v = Vector3DF(*this); v /= op; return v;}
		inline Vector3DF  operator/  (float op) const            {Vector3DF v = Vector3DF(*this); v /= op; return v;}
		inline Vector3DF  operator/  (const Vector3DF &op) const {Vector3DF v = Vector3DF(*this); v /= op; return v;}

		inline Vector3DF &CrossWith (const Vector3DF &op);
		inline Vector3DF  Cross (const Vector3DF &op) const {Vector3DF v = Vector3DF(*this); v.CrossWith(op); return v;}
		
		inline float Dot (const Vector3DF &op) const;
		inline float Dist (const Vector3DF &op) const;
		inline float DistSq (const Vector3DF &op) const;

		inline Vector3DF &Normalize (void);
		inline Vector3DF  Direction (void) const {Vector3DF v = Vector3DF(*this); v.Normalize(); return v;}

		inline float Length (void) const;
		inline float LengthSq (void) const;

		inline VTYPE *Data (void);
	};

	inline Vector3DF operator * (const int op, const Vector3DF &v) {return v * op;}
	inline Vector3DF operator * (const float op, const Vector3DF &v) {return v * op;}
	inline Vector3DF operator / (const int op, const Vector3DF &v) {return Vector3DF((VTYPE)op/v.x, (VTYPE)op/v.y, (VTYPE)op/v.z);}
	inline Vector3DF operator / (const float op, const Vector3DF &v) {return Vector3DF((VTYPE)op/v.x, (VTYPE)op/v.y, (VTYPE)op/v.z);}


	#undef VNAME
	#undef VTYPE

	// Vector4DF Declarations

	#define VNAME		4DF
	#define VTYPE		float

	class Vector4DF {
	public:
		VTYPE x, y, z, w;
	
		// Constructors/Destructors
		inline Vector4DF();
		inline ~Vector4DF();
		inline Vector4DF (VTYPE xa, VTYPE ya, VTYPE za, VTYPE wa);
		inline Vector4DF (const Vector2DI &op);
		inline Vector4DF (const Vector2DF &op);
		inline Vector4DF (const Vector3DI &op);
		inline Vector4DF (const Vector3DF &op);
		inline Vector4DF (const Vector4DF &op);

		// Member Functions
		inline Vector4DF &operator= (const Vector4DF &op);

		inline Vector4DF &operator+= (const Vector4DF &op);
		inline Vector4DF  operator+  (const Vector4DF &op) const {Vector4DF v = Vector4DF(*this); v += op; return v;}

		inline Vector4DF &operator-= (const Vector4DF &op);
		inline Vector4DF  operator-  (const Vector4DF &op) const {Vector4DF v = Vector4DF(*this); v -= op; return v;}

		inline Vector4DF &operator*= (int op);
		inline Vector4DF &operator*= (float op);
		inline Vector4DF &operator*= (const Vector4DF &op);
		Vector4DF &operator*= (MatrixF &op);				// see vector.cpp

		inline Vector4DF  operator*  (int op) const              {Vector4DF v = Vector4DF(*this); v *= op; return v;}
		inline Vector4DF  operator*  (float op) const            {Vector4DF v = Vector4DF(*this); v *= op; return v;}
		inline Vector4DF  operator*  (const Vector4DF &op) const {Vector4DF v = Vector4DF(*this); v *= op; return v;}
		Vector4DF  operator*  (MatrixF &op) const                {Vector4DF v = Vector4DF(*this); v *= op; return v;}

		inline Vector4DF &operator/= (int op);
		inline Vector4DF &operator/= (float op);
		inline Vector4DF &operator/= (const Vector4DF &op);

		inline Vector4DF  operator/  (int op) const              {Vector4DF v = Vector4DF(*this); v /= op; return v;}
		inline Vector4DF  operator/  (float op) const            {Vector4DF v = Vector4DF(*this); v /= op; return v;}
		inline Vector4DF  operator/  (const Vector4DF &op) const {Vector4DF v = Vector4DF(*this); v /= op; return v;}

		Vector4DF &CrossWith (const Vector4DF &op1, const Vector4DF &op2);	// see vector.cpp
		inline Vector3DF  Cross (const Vector4DF &op1, const Vector4DF &op2) const {Vector4DF v = Vector4DF(*this); v.CrossWith(op1,op2); return v;}
		
		inline float Dot (const Vector4DF &op) const;
		inline float Dist (const Vector4DF &op) const;
		inline float DistSq (const Vector4DF &op) const;

		inline Vector4DF &Normalize (void);
		inline Vector4DF  Direction (void) const {Vector4DF v = Vector4DF(*this); v.Normalize(); return v;}

		inline float Length (void) const;
		inline float LengthSq (void) const;

		inline VTYPE *Data (void);
	};
	
	inline Vector4DF operator * (const int op, const Vector4DF &v) {return v * op;}
	inline Vector4DF operator * (const float op, const Vector4DF &v) {return v * op;}
	inline Vector4DF operator / (const int op, const Vector4DF &v) {return Vector4DF((VTYPE)op/v.x, (VTYPE)op/v.y, (VTYPE)op/v.z, (VTYPE)op/v.w);}
	inline Vector4DF operator / (const float op, const Vector4DF &v) {return Vector4DF((VTYPE)op/v.x, (VTYPE)op/v.y, (VTYPE)op/v.z, (VTYPE)op/v.w);}

	#undef VNAME
	#undef VTYPE



	// *********************************
    // Vector Code Definitions (Inlined)
	// *********************************

	// Vector2DI Code Definition

	#define VNAME		2DI
	#define VTYPE		int

	// Constructors/Destructors
	inline Vector2DI::Vector2DI() {x=0; y=0;}
	inline Vector2DI::~Vector2DI() {}
	inline Vector2DI::Vector2DI (VTYPE xa, VTYPE ya) {x=xa; y=ya;}
	inline Vector2DI::Vector2DI (const Vector2DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DI::Vector2DI (const Vector2DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DI::Vector2DI (const Vector3DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DI::Vector2DI (const Vector3DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DI::Vector2DI (const Vector4DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}

	// Member Functions
	inline Vector2DI &Vector2DI::operator= (const Vector2DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; return *this;}

	inline Vector2DI &Vector2DI::operator+= (const Vector2DI &op) {x+=(VTYPE) op.x; y+=(VTYPE) op.y; return *this;}
	inline Vector2DI &Vector2DI::operator-= (const Vector2DI &op) {x-=(VTYPE) op.x; y-=(VTYPE) op.y; return *this;}

	inline Vector2DI &Vector2DI::operator*= (int op) {x*= (VTYPE) op; y*= (VTYPE) op; return *this;}
	inline Vector2DI &Vector2DI::operator*= (float op) {x= (VTYPE)((float)x*op); y= (VTYPE)((float)y*op); return *this;}
	inline Vector2DI &Vector2DI::operator*= (const Vector2DI &op) {x*=(VTYPE) op.x; y*=(VTYPE) op.y; return *this;}

	inline Vector2DI &Vector2DI::operator/= (int op) {x/= (VTYPE) op; y/= (VTYPE) op; return *this;}
	inline Vector2DI &Vector2DI::operator/= (float op) {float inv = 1.0f/op; x= (VTYPE)((float)x*inv); y= (VTYPE)((float)y*inv); return *this;}
	inline Vector2DI &Vector2DI::operator/= (const Vector2DI &op) {x/=(VTYPE) op.x; y/=(VTYPE) op.y; return *this;}

	inline float Vector2DI::Cross (const Vector2DI &v) const	{return (float)y*(float)v.x + (float)(-x)*(float)v.y;}
	inline float Vector2DI::Dot(const Vector2DI &v) const		{return (float) x*v.x + (float) y*v.y;}
	inline float Vector2DI::Dist (const Vector2DI &v) const		{float distsq = DistSq (v); if (distsq>0) return sqrtf(distsq); return 0.0f;}
	inline float Vector2DI::DistSq (const Vector2DI &v) const	{float a,b; a = (float) x - (float) v.x; b = (float) y - (float) v.y; return (a*a + b*b);}

	inline Vector2DI &Vector2DI::Normalize (void) {
		float n = (float) x*x + (float) y*y;
		if (n!=0.0f) {
			n = sqrtf(n);
			x = (VTYPE) (((float) x*255)/n);
			y = (VTYPE) (((float) y*255)/n);				
		}
		return *this;
	}
	inline float Vector2DI::Length (void) const { float n = (float) x*x + (float) y*y; if(n>0) return sqrtf(n); return 0.0f; }
	inline float Vector2DI::LengthSq (void) const { return (float) x*x + (float) y*y; }

	inline VTYPE *Vector2DI::Data (void)			{return &x;}

	#undef VTYPE
	#undef VNAME

	// Vector2DF Code Definition

	#define VNAME		2DF
	#define VTYPE		float

	// Constructors/Destructors
	inline Vector2DF::Vector2DF() {x=0; y=0;}
	inline Vector2DF::~Vector2DF() {}
	inline Vector2DF::Vector2DF (VTYPE xa, VTYPE ya) {x=xa; y=ya;}
	inline Vector2DF::Vector2DF (const Vector2DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DF::Vector2DF (const Vector2DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DF::Vector2DF (const Vector3DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DF::Vector2DF (const Vector3DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}
	inline Vector2DF::Vector2DF (const Vector4DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y;}

	// Member Functions
	inline Vector2DF &Vector2DF::operator= (const Vector2DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; return *this;}

	inline Vector2DF &Vector2DF::operator+= (const Vector2DF &op) {x+=(VTYPE) op.x; y+=(VTYPE) op.y; return *this;}
	inline Vector2DF &Vector2DF::operator-= (const Vector2DF &op) {x-=(VTYPE) op.x; y-=(VTYPE) op.y; return *this;}
		
	inline Vector2DF &Vector2DF::operator*= (int op) {x*= (VTYPE) op; y*= (VTYPE) op; return *this;}
	inline Vector2DF &Vector2DF::operator*= (float op) {x*= (VTYPE) op; y*= (VTYPE) op; return *this;}
	inline Vector2DF &Vector2DF::operator*= (const Vector2DF &op) {x*=(VTYPE) op.x; y*=(VTYPE) op.y; return *this;}

	inline Vector2DF &Vector2DF::operator/= (int op) {float inv = 1.0f/(float)op; x*=inv; y*=inv; return *this;}
	inline Vector2DF &Vector2DF::operator/= (float op) {float inv = 1.0f/op; x*=inv; y*=inv; return *this;}
	inline Vector2DF &Vector2DF::operator/= (const Vector2DF &op) {x/=(VTYPE) op.x; y/=(VTYPE) op.y; return *this;}

	inline float Vector2DF::Cross (const Vector2DF &v) const	{return (float)y*(float)v.x + (float)(-x)*(float)v.y;}
	inline float Vector2DF::Dot(const Vector2DF &v) const		{return (float) x*v.x + (float) y*v.y;}
	inline float Vector2DF::Dist (const Vector2DF &v) const		{float distsq = DistSq (v); if (distsq>0) return sqrtf(distsq); return 0.0f;}
	inline float Vector2DF::DistSq (const Vector2DF &v) const	{float a,b; a = (float) x - (float) v.x; b = (float) y - (float) v.y; return (a*a + b*b);}

	inline Vector2DF &Vector2DF::Normalize (void) {if(x||y) {float invlen = 1.0f/sqrtf(x*x + y*y); x*=invlen; y*=invlen;} return *this;}

	inline float Vector2DF::Length (void) const { float n = (float) x*x + (float) y*y; if(n>0) return sqrtf(n); return 0.0f; }
	inline float Vector2DF::LengthSq (void) const { return (float) x*x + (float) y*y; }

	inline VTYPE *Vector2DF::Data (void)			{return &x;}

	#undef VTYPE
	#undef VNAME

	// Vector3DI Code Definition

	#define VNAME		3DI
	#define VTYPE		int

	// Constructors/Destructors
	inline Vector3DI::Vector3DI() {x=0; y=0; z=0;}
	inline Vector3DI::~Vector3DI() {}
	inline Vector3DI::Vector3DI (VTYPE xa, VTYPE ya, VTYPE za) {x=xa; y=ya; z=za;}
	inline Vector3DI::Vector3DI (const Vector2DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) 0;}
	inline Vector3DI::Vector3DI (const Vector2DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) 0;}
	inline Vector3DI::Vector3DI (const Vector3DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z;}
	inline Vector3DI::Vector3DI (const Vector3DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z;}
	inline Vector3DI::Vector3DI (const Vector4DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z;}

	// Member Functions
	inline Vector3DI &Vector3DI::operator= (const Vector3DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z; return *this;}
	inline Vector3DI &Vector3DI::operator+= (const Vector3DI &op) {x+=(VTYPE) op.x; y+=(VTYPE) op.y; z+=(VTYPE) op.z; return *this;}
	inline Vector3DI &Vector3DI::operator-= (const Vector3DI &op) {x-=(VTYPE) op.x; y-=(VTYPE) op.y; z-=(VTYPE) op.z; return *this;}
		
	inline Vector3DI &Vector3DI::operator*= (int op) {x*= (VTYPE) op; y*= (VTYPE) op; z*= (VTYPE) op; return *this;}
	inline Vector3DI &Vector3DI::operator*= (float op) {x= (VTYPE)((float)x*op); y= (VTYPE)((float)y*op); z= (VTYPE)((float)z*op); return *this;}
	inline Vector3DI &Vector3DI::operator*= (const Vector3DI &op) {x*=(VTYPE) op.x; y*=(VTYPE) op.y; z*=(VTYPE) op.z; return *this;}

	inline Vector3DI &Vector3DI::operator/= (int op) {x/= (VTYPE) op; y/= (VTYPE) op; z/= (VTYPE) op; return *this;}
	inline Vector3DI &Vector3DI::operator/= (float op) {float inv = 1.0f/op; x= (VTYPE)((float)x*inv); y= (VTYPE)((float)y*inv); z= (VTYPE)((float)z*inv); return *this;}
	inline Vector3DI &Vector3DI::operator/= (const Vector3DI &op) {x/=(VTYPE) op.x; y/=(VTYPE) op.y; z/=(VTYPE) op.z; return *this;}

	inline Vector3DI &Vector3DI::CrossWith (const Vector3DI &v) {float ax = (float)x, ay = (float)y, az = (float)z; x = (VTYPE) (ay * (float) v.z - az * (float) v.y); y = (VTYPE) (-ax * (float) v.z + az * (float) v.x); z = (VTYPE) (ax * (float) v.y - ay * (float) v.x); return *this;}
		
	inline float Vector3DI::Dot(const Vector3DI &v) const		{return (float) (x*v.x) + (float) (y*v.y) + (float) (z*v.z);}
	inline float Vector3DI::Dist (const Vector3DI &v) const		{float distsq = DistSq (v); if (distsq>0) return sqrtf(distsq); return 0.0f;}
	inline float Vector3DI::DistSq (const Vector3DI &v) const	{float a,b,c; a = (float) x - (float) v.x; b = (float) y - (float) v.y; c = (float) z - (float) v.z; return (a*a + b*b + c*c);}

	inline Vector3DI &Vector3DI::Normalize (void) {
		float n = (float) x*x + (float) y*y + (float) z*z;
		if (n!=0.0f) {
			n = sqrtf(n);
			x = (VTYPE) (((float) x*255)/n);
			y = (VTYPE) (((float) y*255)/n);
			z = (VTYPE) (((float) z*255)/n);
		}
		return *this;
	}
	inline float Vector3DI::Length (void) const { float n = (float) x*x + (float) y*y + (float) z*z; if(n>0) return sqrtf(n); return 0.0f; }
	inline float Vector3DI::LengthSq (void) const { return (float) x*x + (float) y*y + (float) z*z; }

	inline VTYPE *Vector3DI::Data (void)			{return &x;}

	#undef VTYPE
	#undef VNAME

	// Vector3DF Code Definition

	#define VNAME		3DF
	#define VTYPE		float

	// Constructors/Destructors
	inline Vector3DF::Vector3DF() {x=0; y=0; z=0;}
	inline Vector3DF::~Vector3DF() {}
	inline Vector3DF::Vector3DF (VTYPE xa, VTYPE ya, VTYPE za) {x=xa; y=ya; z=za;}
	inline Vector3DF::Vector3DF (const Vector2DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) 0;}
	inline Vector3DF::Vector3DF (const Vector2DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) 0;}
	inline Vector3DF::Vector3DF (const Vector3DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z;}
	inline Vector3DF::Vector3DF (const Vector3DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z;}
	inline Vector3DF::Vector3DF (const Vector4DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z;}

	// Set Functions
	inline Vector3DF &Vector3DF::Set (float xa, float ya, float za)
	{
		x = xa; y = ya; z = za;
		return *this;
	}

	// Member Functions
	inline Vector3DF &Vector3DF::operator= (const Vector3DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z; return *this;}
	inline Vector3DF &Vector3DF::operator+= (const Vector3DF &op) {x+=(VTYPE) op.x; y+=(VTYPE) op.y; z+=(VTYPE) op.z; return *this;}
	inline Vector3DF &Vector3DF::operator-= (const Vector3DF &op) {x-=(VTYPE) op.x; y-=(VTYPE) op.y; z-=(VTYPE) op.z; return *this;}
		
	inline Vector3DF &Vector3DF::operator*= (int op) {x*= (VTYPE) op; y*= (VTYPE) op; z*= (VTYPE) op; return *this;}
	inline Vector3DF &Vector3DF::operator*= (float op) {x*= (VTYPE) op; y*= (VTYPE) op; z*= (VTYPE) op; return *this;}
	inline Vector3DF &Vector3DF::operator*= (const Vector3DF &op) {x*=(VTYPE) op.x; y*=(VTYPE) op.y; z*=(VTYPE) op.z; return *this;}

	inline Vector3DF &Vector3DF::operator/= (int op) {float inv = 1.0f/(float)op; x*=inv; y*=inv; z*=inv; return *this;}
	inline Vector3DF &Vector3DF::operator/= (float op) {float inv = 1.0f/op; x*=inv; y*=inv; z*=inv; return *this;}
	inline Vector3DF &Vector3DF::operator/= (const Vector3DF &op) {x/=(VTYPE) op.x; y/=(VTYPE) op.y; z/=(VTYPE) op.z; return *this;}

	inline Vector3DF &Vector3DF::CrossWith (const Vector3DF &v) {float ax = x, ay = y; x = (ay * v.z - z * v.y); y = (-ax * v.z + z * v.x); z = (ax * v.y - ay * v.x); return *this;}


	inline float Vector3DF::Dot (const Vector3DF &v) const		{return x*v.x + y*v.y + z*v.z;}
	inline float Vector3DF::Dist (const Vector3DF &v) const		{float distsq = DistSq (v); if (distsq>0) return sqrtf(distsq); return 0.0f;}
	inline float Vector3DF::DistSq (const Vector3DF &v) const	{float a = x - v.x, b = y - v.y, c = z - v.z; return (a*a + b*b + c*c);}

	inline Vector3DF &Vector3DF::Normalize (void) {if(x||y||z) {float invlen = 1.0f/sqrtf(x*x + y*y + z*z); x*=invlen; y*=invlen; z*=invlen;} return *this;}
	inline float Vector3DF::Length (void) const { float n = x*x + y*y + z*z; if(n>0) return sqrtf(n); return 0.0f; }
	inline float Vector3DF::LengthSq (void) const { return x*x + y*y + z*z; }

	inline VTYPE *Vector3DF::Data (void)			{return &x;}

	#undef VTYPE
	#undef VNAME

	// Vector4DF Code Definition

	#define VNAME		4DF
	#define VTYPE		float

	// Constructors/Destructors
	inline Vector4DF::Vector4DF() {x=0; y=0; z=0; w=0;}
	inline Vector4DF::~Vector4DF() {}
	inline Vector4DF::Vector4DF (VTYPE xa, VTYPE ya, VTYPE za, VTYPE wa) {x=xa; y=ya; z=za; w=wa;}
	inline Vector4DF::Vector4DF (const Vector2DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) 0; w=(VTYPE) 0;}
	inline Vector4DF::Vector4DF (const Vector2DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) 0; w=(VTYPE) 0;}
	inline Vector4DF::Vector4DF (const Vector3DI &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z; w=(VTYPE) 0;}
	inline Vector4DF::Vector4DF (const Vector3DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z; w=(VTYPE) 0;}
	inline Vector4DF::Vector4DF (const Vector4DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z; w=(VTYPE) op.w;}

	// Member Functions
	inline Vector4DF &Vector4DF::operator= (const Vector4DF &op) {x=(VTYPE) op.x; y=(VTYPE) op.y; z=(VTYPE) op.z; w=(VTYPE) op.w; return *this;}	
	inline Vector4DF &Vector4DF::operator+= (const Vector4DF &op) {x+=(VTYPE) op.x; y+=(VTYPE) op.y; z+=(VTYPE) op.z; w+=(VTYPE) op.w; return *this;}	
	inline Vector4DF &Vector4DF::operator-= (const Vector4DF &op) {x-=(VTYPE) op.x; y-=(VTYPE) op.y; z-=(VTYPE) op.z; w-=(VTYPE) op.w; return *this;}	

	inline Vector4DF &Vector4DF::operator*= (int op) {x*= (VTYPE) op; y*= (VTYPE) op; z*= (VTYPE) op; w *= (VTYPE) op; return *this;}
	inline Vector4DF &Vector4DF::operator*= (float op) {x*= (VTYPE) op; y*= (VTYPE) op; z*= (VTYPE) op; w *= (VTYPE) op; return *this;}
	inline Vector4DF &Vector4DF::operator*= (const Vector4DF &op) {x*=(VTYPE) op.x; y*=(VTYPE) op.y; z*=(VTYPE) op.z; w*=(VTYPE) op.w; return *this;}	

	inline Vector4DF &Vector4DF::operator/= (int op) {float inv = 1.0f/(float)op; x*=inv; y*=inv; z*=inv; w*=inv; return *this;}
	inline Vector4DF &Vector4DF::operator/= (float op) {float inv = 1.0f/op; x*=inv; y*=inv; z*=inv; w*=inv; return *this;}
	inline Vector4DF &Vector4DF::operator/= (const Vector4DF &op) {x/=(VTYPE) op.x; y/=(VTYPE) op.y; z/=(VTYPE) op.z; w/=(VTYPE) op.w; return *this;}	

	inline float Vector4DF::Dot (const Vector4DF &v) const		{return (float) x*v.x + (float) y*v.y + (float) z*v.z + (float) w*v.w;}
	inline float Vector4DF::Dist (const Vector4DF &v) const		{float distsq = DistSq (v); if (distsq>0) return sqrtf(distsq); return 0.0f;}
	inline float Vector4DF::DistSq (const Vector4DF &v) const	{float a,b,c,d; a = (float) x - (float) v.x; b = (float) y - (float) v.y; c = (float) z - (float) v.z; d = (float) w - (float) v.w; return (a*a + b*b + c*c + d*d);}

	inline Vector4DF &Vector4DF::Normalize (void) {if(x||y||z||w) {float invlen = 1.0f/sqrtf(x*x + y*y + z*z + w*w); x*=invlen; y*=invlen; z*=invlen; w*=invlen;} return *this;}
	inline float Vector4DF::Length (void) const { float n = (float) x*x + (float) y*y + (float) z*z + (float) w*w; if(n>0) return sqrtf(n); return 0.0f; }
	inline float Vector4DF::LengthSq (void) const { return (float) x*x + (float) y*y + (float) z*z + (float) w*w; }

	inline VTYPE *Vector4DF::Data (void)			{return &x;}

	#undef VTYPE
	#undef VNAME

#endif



